//    This file is part of JustMotion.

//    JustMotion is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    JustMotion is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

#include "proc_control.h"

ProcControl::ProcControl(const QString &procFile, const QString &statDir, shared_t *glbShare, QCoreApplication *parent) : QObject(parent)
{
    share         = glbShare;
    fsMon         = new QFileSystemWatcher(this);
    rdThread      = new QThread(nullptr);
    ramDisk       = new RamDisk(rdThread, share->maxMemSize, share->rdMnt, nullptr);
    checkTimer    = new QTimer(this);
    file          = procFile;
    statPath      = statDir;
    closeOnZero   = false;
    procFileAdded = false;
    objCount      = 0;

    checkTimer->setInterval(3000);
    checkTimer->setSingleShot(true);

    connect(fsMon,      &QFileSystemWatcher::fileChanged, this, &ProcControl::procFileUpdated);
    connect(checkTimer, &QTimer::timeout,                 this, &ProcControl::startApp);
}

void ProcControl::createProcFile()
{
    if (!procFileAdded)
    {
        if (!QFileInfo::exists(file))
        {
            QFile fObj(file);

            if (!fObj.open(QFile::WriteOnly))
            {
                QTextStream(stderr) << "err: Failed to open process control file for writing: " << file << " reason: " << fObj.errorString() << Qt::endl;
            }
            else
            {
                fObj.write("##");
                fObj.close();

                procFileAdded = fsMon->addPath(file);
            }
        }
        else
        {
            QTextStream(stdout) << "wrn: " << file << " already exists so it will be removed, this will cause any other existing instance to close." << Qt::endl;

            if (!QFile::remove(file))
            {
                QTextStream(stderr) << "err: Failed to remove process control file: " << file << " check permissions." << Qt::endl;
            }
        }
    }
}

void ProcControl::startRamdisk()
{
    if (!rdThread->isRunning())
    {
        rdThread->start();
    }
}

void ProcControl::umntThenClose()
{
    QDir(statPath).removeRecursively();

    if (rdThread->isRunning())
    {
        for (int i = 0; i < 3; ++i)
        {
            thread()->sleep(1);

            if (QProcess::execute("umount", {share->rdMnt}) == 0)
            {
                break;
            }
        }
    }

    QCoreApplication::instance()->quit();
}

void ProcControl::createCamObjs()
{
    if (procFileAdded)
    {
        for (auto &&conf : lsFilesInDir(share->confPath))
        {
            auto confFilePath = QDir::cleanPath(share->confPath) + "/" + conf;

            if (!confList.contains(confFilePath))
            {
                new Camera(*share, confFilePath, statPath, ramDisk, this, QCoreApplication::instance());
            }
        }
    }
}

void ProcControl::procFileUpdated(const QString &path)
{
    fsMon->removePath(path);

    QFile::remove(path);

    closeApp();
}

void ProcControl::closeApp()
{
    if (confList.isEmpty())
    {
        umntThenClose();
    }
    else
    {
        closeOnZero = true;

        emit prepForClose();
    }
}

void ProcControl::startApp()
{
    if (!closeOnZero)
    {
        startRamdisk();
        createProcFile();
        createCamObjs();

        checkTimer->start();
    }
}

void ProcControl::objPlusOne(const QString &conf)
{
    confList.append(conf);
}

void ProcControl::objMinusOne(const QString &conf)
{
    confList.removeAll(conf);

    if (closeOnZero && confList.isEmpty())
    {
        umntThenClose();
    }
}
