//    This file is part of JustMotion.

//    JustMotion is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    JustMotion is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

#include "ram_disk.h"

mode_t fileDefaultMode()
{
    return S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH | S_IFREG;
}

mode_t symmDefaultMode()
{
    return S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH | S_IFLNK;
}

mode_t dirDefaultMode()
{
    return S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH | S_IFDIR;
}

RamDisk                *RamDisk::instance  = nullptr;
QString                 RamDisk::mnt       = QString();
QStringList             RamDisk::flushHist = QStringList();
QStringList             RamDisk::bondDirs  = QStringList();
QHash<QString, file_t*> RamDisk::data      = QHash<QString, file_t*>();
fuse_operations         RamDisk::fs_oper   = fuse_operations();
quint64                 RamDisk::inoSeed   = 2;
quint64                 RamDisk::rootCtime = 0;
quint64                 RamDisk::rootMtime = 0;
quint64                 RamDisk::duMax     = 0;
quint64                 RamDisk::du        = 0;
QRecursiveMutex         RamDisk::mutex     = QRecursiveMutex();

RamDisk::RamDisk(QThread *thr, quint64 maxSize, const QString &mntPnt, QObject *parent) : QObject(parent)
{
    duMax     = maxSize;
    mnt       = mntPnt;
    instance  = this;
    rootCtime = QDateTime::currentMSecsSinceEpoch();
    rootMtime = rootCtime;

    connect(thr, &QThread::started,  this, &RamDisk::exec);

    moveToThread(thr);
}

void RamDisk::lock()
{
    mutex.lock();
}

void RamDisk::unlock()
{
    mutex.unlock();
}

void RamDisk::exec()
{
    fs_oper.getattr	        = RamDisk::getattr;
    fs_oper.readlink        = RamDisk::readlink;
    fs_oper.readdir	        = RamDisk::readdir;
    fs_oper.mkdir	        = RamDisk::mkdir;
    fs_oper.symlink	        = RamDisk::symlink;
    fs_oper.unlink	        = RamDisk::unlink;
    fs_oper.rmdir	        = RamDisk::rmdir;
    fs_oper.rename	        = RamDisk::rename;
    fs_oper.truncate        = RamDisk::truncate;
    fs_oper.open	        = RamDisk::open;
    fs_oper.create 	        = RamDisk::create;
    fs_oper.read	        = RamDisk::read;
    fs_oper.write	        = RamDisk::write;
    fs_oper.statfs	        = RamDisk::statfs;
    fs_oper.init            = RamDisk::init;
    fs_oper.destroy         = RamDisk::destroy;
    fs_oper.utimens         = RamDisk::utimens;
    fs_oper.fallocate       = RamDisk::fallocate;
    fs_oper.flush           = RamDisk::flush;
    fs_oper.release	        = RamDisk::release;
    fs_oper.chmod	        = RamDisk::chmod;
    fs_oper.chown           = RamDisk::chown;
    fs_oper.copy_file_range = NULL;
    fs_oper.link	        = NULL;
    fs_oper.lock            = NULL;
    fs_oper.releasedir      = NULL;
    fs_oper.fsync	        = NULL;
    fs_oper.opendir         = NULL;
    fs_oper.fsyncdir        = NULL;
    fs_oper.bmap            = NULL;
    fs_oper.ioctl           = NULL;
    fs_oper.flock           = NULL;
    fs_oper.lseek           = NULL;
    fs_oper.access          = NULL;
    fs_oper.write_buf       = NULL;
    fs_oper.read_buf        = NULL;
    fs_oper.getxattr        = NULL;
    fs_oper.setxattr        = NULL;
    fs_oper.listxattr       = NULL;
    fs_oper.removexattr     = NULL;

    QByteArray appName(APP_TARGET);
    //QByteArray debug("-d");
    //QByteArray singleTh("-s");
    QByteArray foreground("-f");
    QByteArray option("-o");
    //QByteArray dp("default_permissions");
    QByteArray other("allow_other");

    int argc = 5;

    char *argv[] = {appName.data(),
                    //debug.data(),
                    //singleTh.data(),
                    foreground.data(),
                    //option.data(),
                    //dp.data(),
                    option.data(),
                    other.data(),
                    mnt.toUtf8().data()};

    auto ret = fuse_main(argc, argv, &fs_oper, NULL);

    if (ret != 0)
    {
        qCritical() << "err: the ram disk failed to mount, return code: " << ret;
    }
}

bool RamDisk::isFull()
{
    return du >= duMax;
}

int RamDisk::calcNumOfBlocks(quint64 len, bool countZero)
{
    if ((len == 0) && !countZero)
    {
        return 0;
    }
    if (len == 0)
    {
        return 1;
    }
    else
    {
        qreal blks = static_cast<qreal>(len) / static_cast<qreal>(BLK_SIZE);

        return static_cast<int>(ceil(blks));
    }
}

void RamDisk::fillStatTimes(struct stat *info, quint64 cTimeSec, quint64 mTimeSec)
{
    // this converts the UTC epoch stored in the ram disk to
    // local epoch using QDateTime.
    auto cTime = QDateTime::fromMSecsSinceEpoch(cTimeSec, Qt::LocalTime);
    auto mTime = QDateTime::fromMSecsSinceEpoch(mTimeSec, Qt::LocalTime);
    auto curr  = QDateTime::currentDateTime();

    info->st_ctim.tv_sec = cTime.toSecsSinceEpoch();
    info->st_mtim.tv_sec = mTime.toSecsSinceEpoch();

    info->st_ctim.tv_nsec = cTime.toMSecsSinceEpoch();
    info->st_mtim.tv_nsec = mTime.toMSecsSinceEpoch();

    info->st_atim.tv_nsec = curr.toMSecsSinceEpoch();
    info->st_atim.tv_sec  = curr.toSecsSinceEpoch();
}

int RamDisk::getattr(const char *path, struct stat *info, struct fuse_file_info *fi)
{
    Q_UNUSED(fi);

    lock();

    auto cleanedPath = QDir::cleanPath(path);
    auto ret         = 0;

    memset(info, 0, sizeof(struct stat));

    info->st_uid     = getuid();
    info->st_gid     = getgid();
    info->st_nlink   = 1;
    info->st_blksize = BLK_SIZE;

    if (cleanedPath == "/")
    {
        info->st_mode   = dirDefaultMode();
        info->st_ino    = 1;
        info->st_size   = 0;
        info->st_blocks = 0;

        fillStatTimes(info, rootCtime, rootMtime);
    }
    else if (data.contains(cleanedPath))
    {
        auto file = data[cleanedPath];

        info->st_ino    = file->ino;
        info->st_mode   = file->mode;
        info->st_size   = file->bytes.size();
        info->st_blocks = file->blks;

        fillStatTimes(info, file->ctime, file->mtime);
    }
    else
    {
        ret = -ENOENT;
    }

    unlock();

    return ret;
}

int RamDisk::readlink(const char *path, char *buff, size_t len)
{
    auto cleanedPath = QDir::cleanPath(path);
    auto ret         = 0;

    lock();

    if (data.contains(cleanedPath))
    {
        memcpy(buff, data[cleanedPath]->bytes.data(), len);
    }
    else
    {
        ret = -ENOENT;
    }

    unlock();

    return ret;
}

QString RamDisk::getFileName(const QString &basePath, const QString &fullPath)
{
    QString ret;

    if (fullPath.startsWith(basePath))
    {
        auto name = fullPath.mid(basePath.size());

        if (name.startsWith("/"))
        {
            name = name.mid(1);
        }

        if (name.indexOf("/") == -1)
        {
            ret = name;
        }
    }

    return ret;
}

bool RamDisk::dirEmpty(const QString &path, int &ret)
{
    ret = 0;

    if (!data.contains(path) && path != "/")
    {
        ret = -ENOENT;
    }
    else
    {
        auto treeList = data.keys();

        for (auto fullPath : treeList)
        {
            auto name = getFileName(path, fullPath);

            if (!name.isEmpty())
            {
                ret = -ENOTEMPTY; break;
            }
        }
    }

    return ret == 0;
}

quint32 RamDisk::getMode(const QString &path)
{
    quint32 ret = 0;

    if (path == "/")
    {
        ret = dirDefaultMode();
    }
    else if (data.contains(path))
    {
        ret = data[path]->mode;
    }

    return ret;
}

int RamDisk::readdir(const char *path, void *buff, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi, enum fuse_readdir_flags flags)
{
    Q_UNUSED(offset);
    Q_UNUSED(fi);
    Q_UNUSED(flags);

    lock();

    auto fileList    = data.keys();
    auto cleanedPath = QDir::cleanPath(path);
    auto ret         = 0;

    struct stat st;

    memset(&st, 0, sizeof(st));

    for (auto filePath : fileList)
    {
        auto name = getFileName(cleanedPath, filePath);

        if (!name.isEmpty())
        {
            auto file = data[QDir::cleanPath(cleanedPath + "/" + name)];

            st.st_mode   = file->mode;
            st.st_ino    = file->ino;
            st.st_size   = file->bytes.size();
            st.st_blocks = file->blks;

            fillStatTimes(&st, file->ctime, file->mtime);

            if (filler(buff, name.toUtf8().data(), &st, 0, FUSE_FILL_DIR_PLUS))
            {
                ret = -ENOMEM;
            }
        }
    }

    unlock();

    return ret;
}

int RamDisk::mkdir(const char *path, mode_t mode)
{
    Q_UNUSED(mode);

    lock();

    auto cleanedPath = QDir::cleanPath(path);
    auto ret         = 0;

    if (!data.contains(cleanedPath) && (cleanedPath != "/"))
    {
        auto info = new file_t();
        auto curr = QDateTime::currentDateTimeUtc();

        info->mode  = dirDefaultMode();
        info->ino   = inoSeed++;
        info->blks  = 0;
        info->mtime = curr.toMSecsSinceEpoch();
        info->ctime = curr.toMSecsSinceEpoch();

        data.insert(cleanedPath, info);
    }

    unlock();

    return ret;
}

int RamDisk::symlink(const char *src, const char *dst)
{
    lock();

    auto cleanedSrcPath = QDir::cleanPath(src);
    auto cleanedDstPath = QDir::cleanPath(dst);
    auto ret            = 0;

    if (isFull())
    {
        ret = -ENOSPC;
    }
    else
    {
        auto info = new file_t();
        auto curr = QDateTime::currentDateTimeUtc();

        info->mode  = symmDefaultMode();
        info->ino   = inoSeed++;
        info->blks  = 1;
        info->bytes = cleanedSrcPath.toUtf8();
        info->mtime = curr.toMSecsSinceEpoch();
        info->ctime = curr.toMSecsSinceEpoch();

        du += info->bytes.size();

        data.insert(cleanedDstPath, info);
    }

    unlock();

    return ret;
}

int RamDisk::unlink(const char *path)
{
    lock();

    auto cleanedPath = QDir::cleanPath(path);
    auto ret         = 0;

    if (!data.contains(cleanedPath))
    {
        ret = -ENOENT;
    }
    else
    {
        du -= data[cleanedPath]->bytes.size();

        data.remove(cleanedPath);
    }

    unlock();

    return ret;
}

int RamDisk::rmdir(const char *path)
{
    lock();

    auto cleanedPath = QDir::cleanPath(path);
    auto ret         = 0;

    if (dirEmpty(cleanedPath, ret))
    {
        data.remove(cleanedPath);
    }

    unlock();

    return ret;
}

void RamDisk::rePath(const QString &src, const QString &dst)
{
    auto pathList = data.keys();

    for (auto path : pathList)
    {
        if (path.startsWith(src))
        {
            auto file    = data.take(path);
            auto newPath = path.replace(src, dst);

            data.insert(newPath, file);
        }
    }
}

int RamDisk::rename(const char *src, const char *dst, unsigned int flags)
{
    lock();

    auto cleanedSrcPath = QDir::cleanPath(src);
    auto cleanedDstPath = QDir::cleanPath(dst);
    auto srcMode        = getMode(cleanedSrcPath);
    auto dstMode        = getMode(cleanedDstPath);
    auto ret            = 0;

    if (srcMode && dstMode)
    {
        if (srcMode != dstMode)
        {
            ret = -EINVAL;
        }
        else if (flags & RENAME_EXCHANGE)
        {
            auto tmp = QString(TMP_DIR) + QDir::separator() + QString(cleanedSrcPath);

            rePath(cleanedSrcPath, tmp);
            rePath(cleanedDstPath, cleanedSrcPath);
            rePath(tmp, cleanedDstPath);
        }
        else if ((flags & RENAME_NOREPLACE) == 0)
        {
            rePath(cleanedSrcPath, cleanedDstPath);
        }
        else
        {
            ret = -EEXIST;
        }
    }
    else if (srcMode && !dstMode)
    {
        rePath(cleanedSrcPath, cleanedDstPath);
    }
    else
    {
        ret = -ENOENT;
    }

    unlock();

    return ret;
}

int RamDisk::open(const char *path, struct fuse_file_info *fi)
{
    Q_UNUSED(path);
    Q_UNUSED(fi);

    return 0;
}

int RamDisk::truncate(const char *path, off_t size, struct fuse_file_info *fi)
{
    Q_UNUSED(fi);

    lock();

    auto cleanedPath = QDir::cleanPath(path);
    auto ret         = 0;

    if (!data.contains(cleanedPath))
    {
        ret = -ENOENT;
    }
    else
    {
        auto file    = data[cleanedPath];
        auto oldSize = file->bytes.size();

        file->bytes = file->bytes.leftJustified(size, 0x00, true);
        file->mtime = QDateTime::currentDateTimeUtc().toMSecsSinceEpoch();

        if (oldSize > size)
        {
            du -= oldSize - size;
        }
        else if (size > oldSize)
        {
            du += size - oldSize;
        }

        file->blks = calcNumOfBlocks(size);
    }

    unlock();

    return ret;
}

int RamDisk::create(const char *path, mode_t mode, struct fuse_file_info *fi)
{
    Q_UNUSED(mode);
    Q_UNUSED(fi);

    lock();

    auto cleanedPath = QDir::cleanPath(path);
    auto ret         = 0;

    if (cleanedPath == "/")
    {
        ret = -EINVAL;
    }
    else if (data.contains(cleanedPath))
    {
        ret = -EEXIST;
    }
    else
    {
        auto file = new file_t();
        auto curr = QDateTime::currentDateTimeUtc();

        file->ino   = inoSeed++;
        file->blks  = 1;
        file->mode  = fileDefaultMode();
        file->ctime = curr.toMSecsSinceEpoch();
        file->mtime = curr.toMSecsSinceEpoch();

        data.insert(cleanedPath, file);
    }

    unlock();

    return ret;
}

int RamDisk::read(const char *path, char *buff, size_t len, off_t offs, struct fuse_file_info *fi)
{
    Q_UNUSED(fi);

    lock();

    auto cleanedPath = QDir::cleanPath(path);
    auto ret         = 0;

    if (!data.contains(cleanedPath))
    {
        ret = -ENOENT;
    }
    else
    {
        auto file  = data[cleanedPath];
        auto iOffs = static_cast<int>(offs);
        auto iLen  = static_cast<int>(len);

        if (iOffs > (file->bytes.size() - 1))
        {
            ret = -EFAULT;
        }
        else
        {
            if ((iOffs + iLen) > file->bytes.size())
            {
                iLen = file->bytes.size() - iOffs;
            }

            memcpy(buff, file->bytes.data() + iOffs, iLen);

            ret = iLen;
        }
    }

    unlock();

    return ret;
}

int RamDisk::write(const char *path, const char *buff, size_t len, off_t offs, struct fuse_file_info *fi)
{
    Q_UNUSED(fi);

    lock();

    auto cleanedPath = QDir::cleanPath(path);
    auto ret         = 0;

    if (isFull())
    {
        ret = -ENOSPC;
    }
    else if (!data.contains(cleanedPath))
    {
        ret = -ENOENT;
    }
    else
    {
        auto file    = data[cleanedPath];
        auto oldSize = file->bytes.size();

        file->bytes.replace(offs, len, QByteArray::fromRawData(buff, len));

        if (oldSize > file->bytes.size())
        {
            du -= oldSize - file->bytes.size();
        }
        else if (file->bytes.size() > oldSize)
        {
            du += file->bytes.size() - oldSize;
        }

        file->blks = calcNumOfBlocks(file->bytes.size());

        ret = len;
    }

    unlock();

    return ret;
}

int RamDisk::statfs(const char *path, struct statvfs *stbuff)
{
    Q_UNUSED(path);

    stbuff->f_frsize = BLK_SIZE;
    stbuff->f_bsize  = BLK_SIZE;
    stbuff->f_blocks = calcNumOfBlocks(duMax);
    stbuff->f_bfree  = stbuff->f_blocks - calcNumOfBlocks(du, false);
    stbuff->f_bavail = stbuff->f_bfree;

    return 0;
}

void RamDisk::destroy(void *privateData)
{
    Q_UNUSED(privateData);

    lock();

    data.clear();

    unlock();

    inoSeed = 1;
    du      = 0;
}

int RamDisk::utimens(const char *path, const struct timespec newTimes[2], struct fuse_file_info *fi)
{
    Q_UNUSED(fi);
    Q_UNUSED(newTimes);

    lock();

    auto cleanedPath = QDir::cleanPath(path);
    auto ret         = 0;

    if (!data.contains(cleanedPath))
    {
        ret = -ENOENT;
    }
    else
    {
        data[cleanedPath]->mtime = QDateTime::currentDateTimeUtc().toMSecsSinceEpoch();
    }

    unlock();

    return ret;
}

int RamDisk::fallocate(const char *path, int mode, off_t offs, off_t len, struct fuse_file_info *fi)
{
    Q_UNUSED(path);
    Q_UNUSED(mode);
    Q_UNUSED(fi);
    Q_UNUSED(offs);
    Q_UNUSED(len);

    return 0;
}

int RamDisk::flush(const char *path, struct fuse_file_info *fi)
{
    Q_UNUSED(fi);

    auto cleanedPath = QDir::cleanPath(path);
    auto basePath    = QFileInfo(cleanedPath).path();

    if (flushHist.size() >= (bondDirs.size() * 4))
    {
        flushHist = flushHist.mid(bondDirs.size() * 2);
    }

    if (bondDirs.contains(basePath) && !flushHist.contains(cleanedPath))
    {
        flushHist.append(cleanedPath);

        emit instance->watchedFileFlushed(cleanedPath);
    }

    return 0;
}

int RamDisk::release(const char *path, struct fuse_file_info *fi)
{
    Q_UNUSED(path);
    Q_UNUSED(fi);

    return 0;
}

int RamDisk::chmod(const char *path, mode_t mode, struct fuse_file_info *fi)
{
    Q_UNUSED(path);
    Q_UNUSED(mode);
    Q_UNUSED(fi);

    return 0;
}

int RamDisk::chown(const char *path, uid_t uid, gid_t gid, struct fuse_file_info *fi)
{
    Q_UNUSED(path);
    Q_UNUSED(uid);
    Q_UNUSED(gid);
    Q_UNUSED(fi);

    return 0;
}

void *RamDisk::init(struct fuse_conn_info *conn, struct fuse_config *cfg)
{
    Q_UNUSED(conn);
    Q_UNUSED(cfg);

    return NULL;
}

void RamDisk::addCam(const QString &name)
{
    bondDirs.append("/" + name + "/vid");
}

void RamDisk::rmCam(const QString &name)
{
    bondDirs.removeAll("/" + name + "/vid");
}
